$COM 0
$COMPILE UNIT "SIMPLEMS.PBU"
$CPU 8086
$DEBUG MAP -,PBDEBUG -,PATH -,UNIT -
$DIM ALL
$ERROR ALL -
$EVENT -
$FLOAT EMULATE
$LIB ALL -
$OPTIMIZE SPEED
$OPTION CNTLBREAK -,GOSUB -,SIGNED -
$SOUND 1
$STACK 2048
$STRING 1
DECLARE FUNCTION EMSReady () AS BYTE
DECLARE FUNCTION EMSVersion () AS BYTE
DECLARE FUNCTION EMSPGFrame () AS WORD
DECLARE FUNCTION EMSTotMem () AS DWORD
DECLARE FUNCTION EMSAveMem () AS DWORD
DECLARE FUNCTION EMSStatus () AS BYTE
DECLARE FUNCTION EMSAlloc (NPages AS WORD) AS WORD
DECLARE FUNCTION EMSDealloc (EHandle AS WORD) AS BYTE
DECLARE FUNCTION EMSMapMem (EHandle AS WORD,YPage AS WORD,TPage AS BYTE) AS BYTE
DECLARE FUNCTION EMSUnMap (TPage AS BYTE) AS BYTE
DECLARE FUNCTION EMSSave (EHandle AS WORD) AS BYTE
DECLARE FUNCTION EMSLoad (EHandle AS WORD) AS BYTE

'The SimpleEMS v1.0 source code. This source is meant to be compiled with
'PB v3.2, as a PBU (PowerBASIC Unit).

'SimpleEMS v1.0 by Ray, from Sureal.
'E-mail: R.Zwitserloot@BTInternet.com

'given to the public domain. I frankly don't care what you want to do with
'it. Just don't try and hold me responsible or something.. I give you the
'source on an AS IS basis, blah, blah, blah... I don't really have to bother
'with legalities when supplying Public Domain, I hope.

'All subbies and functions are remarked, and explained. There is even
'a .PBH and the source for that to implement straight into your PB
'editor, so you can see what these things do as you program :-).

'aint I a nice guy.
'NB: PageDOWN on.. every new sub/func should fall on a new page.
'NB2: I like assembler. This should be quite obvious.
'NB3: these tables are all over the place. This was really simple.
'--------------------------------------------------------------

%FALSE = 0             'The usual equates.
%TRUE = NOT %FALSE

'-----------------------------------------------------------------------
'FUNCTION EMSReady
'inputs: none.
'outputs: 0 (FALSE) if it aint there. -1 (TRUE) if it is.  (BYTE)
'-----------------------------------------------------------------------
'EXPLANATION:
'your EMM driver installs itself as a device, named EMMXXXX0. kinda like,
'say, COM1. you can open these things as files. So, if you try to open
'this files, and dos says it aint there, EMM isn't there or got lost
'somewhere. It could of course be a real file, so then you use IOCTL to
'check if the thing is ready for output. It should be. If not, EMM is
'in trouble, or you actually ran over a FILE called EMMXXXX0. rather than
'the devicedriver. I still don't get PB's commands for IOCTL, so I just
'used Good Ole ASM.
FUNCTION EMSReady () STATIC PUBLIC AS BYTE
DIM EMSName AS STRING*9
DIM EMSP AS WORD
EMSP=VARPTR(EMSName)
EMSName="EMMXXXX0"+CHR$(0)  'dos only eats strings ending in a NULL.
IF DIR$("EMMXXXX0")="" THEN EMSReady=%FALSE:EXIT FUNCTION
'Opening file, DOS interrupt method. That's how PB does it,
'exept PB uses an internal handle system. TIP to Bob:
'how about letting us know the real handle by some function?
'see, in this case, I needs to know my handle so that I can call
'IOCTL and query this particular device.
!MOV AX,&h3D00  ;dos func open file. (3D)
!MOV DX,EMSP    ;load adress of name terminated with NULL in DX.
!INT &h21       ;go.
!JC EMSRFalse1  ;carry bit set- file doenst exit. exit.
!MOV BX,AX      ;handle returned in AX. subsequent calls want it in BX.
!MOV AX,&h4407  ;dos func IOCTL(44)/sub func get output status (7)
!INT &h21       ;go.
!JC EMSRFalse2  ;no IOCTL support. exit&close file.
!CMP AL,&h0FF   ;check for readyness.
!JNZ EMSRFalse2 ;not ready for output-not EMM.
!MOV AH,&h3E    ;dos func-close file. (3E)
!INT &h21       ;go.
EMSReady=%TRUE  'it was there.
EXIT FUNCTION
EMSRFalse2:
!MOV AH,&h3E    ;it wasn't there. close the file.
!INT &h21
EMSRFalse1:
EMSReady=%FALSE
END FUNCTION


'-----------------------------------------------------------------------
'FUNCTION EMSVersion
'inputs: none.
'outputs: version. (NB: version 4.2 would be 42). (BYTE)
'-----------------------------------------------------------------------
'EXPLANATION:
'after you are sure EMM is installed, you can use interrupt &h67 to talk
'with EMM. If EMM doesn't exist, you are likely to crash your computer.
'That's because in that case there will be a 0000:0000 in the interrupt
'vector table for INT &h67, and so, when you call int 67, it will begin
'executing code at that adress. the interrupt table resides here. 'Running'
'data like that is a surefire way to crash and burn :-). Never use
'any EMS functions before making sure it's there with EMSReady...
'SimplEMS functions don't check for it, except, of course, EMSReady.
'EMS func- 46h, get version.
'note on the version returned: if you get, say, &h43, it's version 4.3
'NB: If you are used to ASM programming.. EMS returns version as BCD-HEX.
'EMSVersion returns as decimal. ie: if you get back 43 (decimal), that
'would mean you have version 4.3.
'NB: PB has a bug here. If you don't use EMSV as stepstone, and use
'EMSVersion right away, you get a fatal crash. Even:
'EMSVersion=VAL(HEX$(EMSVersion)) causes this to happen. anyway, I
'have reported this.

FUNCTION EMSVersion () STATIC PUBLIC AS BYTE
DIM EMSV AS BYTE
!MOV AH,&h46
!INT &h67
!MOV EMSV,AL
EMSVersion=VAL(HEX$(EMSV))
END FUNCTION



'-----------------------------------------------------------------------
'FUNCTION EMSPGFrame
'inputs: none.
'outputs: Active Page Frame Segment. (WORD)
'-----------------------------------------------------------------------
'EXPLANATION:
'See how EMS operates as to what the APFS is. It uses EMS func 41.

FUNCTION EMSPGFrame () STATIC PUBLIC AS WORD
DIM EMSPGT AS WORD
!MOV AH,&h41
!INT &h67
!MOV EMSPGT,BX
EMSPGFrame=EMSPGT
END FUNCTION


'-----------------------------------------------------------------------
'FUNCTION EMSTotMem
'inputs: none.
'outputs: Total EMS Memory.   (DWORD)
'notes: use CDBL(EMSTotMem) to avoid PB reporting in Scientific notation.
'-----------------------------------------------------------------------
'EXPLANATION:
'uses EMS func &h42 to get the total EMS memory. That might all be in use,
'though. use EMSAveMem to find available (free) EMS mem.
'using this function returns Total pages in DX, and available pages in BX.
'(IE: both EMSTotMem and EMSAveMem call same EMS func.) one page is 16k,
'so we multiply that by &h4000 (16k) to get to our result.

FUNCTION EMSTotMem () STATIC PUBLIC AS DWORD
DIM EMST AS DWORD
DIM EMST2 AS WORD
!MOV AH,&h42
!INT &h67
!MOV EMST2,DX
EMST=EMST2*&h4000  'multiply #pages with 16k.
EMSTotMem=EMST
END FUNCTION

'-----------------------------------------------------------------------
'FUNCTION EMSAveMem
'inputs: none.
'outputs: Available EMS Memory.   (DWORD)
'notes: use CDBL(EMSAveMem) to avoid PB reporting in Scientific notation.
'-----------------------------------------------------------------------
'EXPLANATION:
'uses EMS func &h42 to get the Ave EMS memory. You might have more used,
'though. use EMSTotMem to find total (free+used) EMS mem.
'using this function returns Total pages in DX, and available pages in BX.
'(IE: both EMSTotMem and EMSAveMem call same EMS func.) one page is 16k,
'so we multiply that by &h4000 (16k) to get to our result.

FUNCTION EMSAveMem () STATIC PUBLIC AS DWORD
DIM EMST AS DWORD
DIM EMST2 AS WORD
!MOV AH,&h42
!INT &h67
!MOV EMST2,BX
EMST=EMST2*&h4000  'multiply #pages with 16k.
EMSAveMem=EMST
END FUNCTION




'-----------------------------------------------------------------------
'FUNCTION EMSStatus
'inputs: none.
'outputs: TRUE (-1) if EMS is functional, false (0) if it isn't
'-----------------------------------------------------------------------
'EXPLANATION:
'uses EMS func 40.

FUNCTION EMSStatus () STATIC PUBLIC AS BYTE
DIM Tmp AS BYTE
!MOV AH,&h40
!INT &h67
!MOV Tmp,AH
EMSStatus=%FALSE
IF Tmp=0 THEN EMSStatus=%TRUE
END FUNCTION


'-----------------------------------------------------------------------
'FUNCTION EMSAlloc
'inputs: amount of 16k pages to alloc. (WORD)
'outputs: Handle. if handle is &hFFFF, something went wrong. (WORD)
'notes: take care to release allocated mem with EMSDealloc, lest you
'make that EMS memory unavailable until reboot. This function does
'not allow you to access your memory, merely claim some.
'-----------------------------------------------------------------------
'EXPLANATION:
'using function 43, alloc mem, we reserve some ems memory.

FUNCTION EMSAlloc (NPages AS WORD) STATIC PUBLIC AS WORD
DIM EHandle AS WORD
DIM NP AS WORD
NP=NPages
!MOV AH,&h43
!MOV BX,NP
!INT &h67
!MOV EHandle,DX
!CMP AH,0
!JNZ EMSAErrorInAlloc
EMSAlloc=EHandle
EXIT FUNCTION

EMSAErrorInAlloc:
EMSAlloc=&hFFFF
END FUNCTION

'-----------------------------------------------------------------------
'FUNCTION EMSDealloc
'inputs: Handle of memory to release. (WORD)
'outputs: true (-1) if it worked out, false (0) if it didnt. (BYTE)
'notes: use to let go of memory claimed with EMSAlloc
'-----------------------------------------------------------------------
'EXPLANATION:
'using function 45, release mem, we let go of what we claimed
'with EMSAlloc

FUNCTION EMSDealloc (EHandle AS WORD) STATIC PUBLIC AS BYTE
DIM EH AS WORD
EH=EHandle
!MOV AH,&h45
!MOV DX,EH
!INT &h67
!CMP AH,0
!JNZ EMSDErrorInDealloc
EMSDealloc=%TRUE
EXIT FUNCTION
EMSDErrorInDealloc:
EMSDealloc=%FALSE
END FUNCTION



'-----------------------------------------------------------------------
'FUNCTION EMSMapMem
'inputs: Handle (WORD), Your Page (WORD), Target Page (BYTE).
'outputs: true (-1) if it worked out, false (0) if it didnt. (BYTE)
'notes: use to access your memory.
'-----------------------------------------------------------------------
'EXPLANATION:
'To actually access your memory, you give the handle (gotten with
'EMSAlloc) to this routine, along with the page within your allocated
'memory (ie: if you allocated 200 pages, and you want to access the 34th
'16k block, load the second parameter with 34.), and as third parameter,
'give the target 16k block in real mem. this is a value between 1 and 4.
'See the help for more info.
'NB: note that in asm-speak, the first page would be 0. here, it would be 1.
'uses EMS func 44, map memory.

FUNCTION EMSMapMem (EHandle AS WORD,YPage AS WORD,TPage AS BYTE) STATIC PUBLIC AS BYTE
DIM EH AS WORD
DIM YP AS WORD
DIM TP AS BYTE
EH=EHandle
YP=YPage
TP=TPage
DECR YP
DECR TP
!MOV AH,&h44
!MOV DX,EH
!MOV BX,YP
!MOV AL,TP
!INT &h67
!CMP AH,0
!JNZ EMSMMError
EMSMapMem=%TRUE
EXIT FUNCTION
EMSMMError:
EMSMapMem=%FALSE
END FUNCTION



'-----------------------------------------------------------------------
'FUNCTION EMSUnMap
'inputs: Target Page (BYTE).
'outputs: true (-1) if it worked out, false (0) if it didnt. (BYTE)
'notes: useless, unless used in conjunction with EMSSave and EMSLoad.
'       this function merely tells the EMM that you don't need the
'       memory there on future calls. Even then, if you don't use this
'       function, another program will map right over yours. It only
'       comes into play with running external programs messing with EMS.
'       (TSRs.), and then only when you are writing one.
'-----------------------------------------------------------------------
'EXPLANATION:
'see EMSMapMem. When you load handle with -1, (&hFFFF) it will unmap
'memory from the target page, making it unavailable until remapped with
'EMSMapMem. Doesn't change contents of the actual Target Map, merely
'in the internal registers for EMSSave and EMSLoad.

FUNCTION EMSUnMap (TPage AS BYTE) STATIC PUBLIC AS BYTE
DIM TP AS BYTE
TP=TPage
DECR TP
!MOV AH,&h44
!MOV BX,TP
!INT &h67
!CMP AH,0
!JNZ EMSUMError
EMSUnMap=%TRUE
EXIT FUNCTION
EMSUMError:
EMSUnMap=%FALSE
END FUNCTION



'-----------------------------------------------------------------------
'FUNCTION EMSSave
'inputs: Your EMS Handle. (WORD)
'outputs: true (-1) if it worked out, false (0) if it didnt. (BYTE)
'notes: If you are interrupting some program that uses EMS, and you
'       map, you are potentially screwing things up for the other
'       program. the operating system takes care of stuff like that
'       for itself, but other programs can't. When using virtual
'       machines or multithreading this won't be a problem either, but
'       it WILL be a problem when writing TSRs, particularly active ones.
'       (pops up on key-combos, timer, etc, rather than interrupt/multiplex)
'       In that case, you can save the status of the memory maps before
'       you begin mapping, and restore it when you exit. Supply your handle
'       so the EMM can support multiple saves. NB: use EMSSave before you
'       begin EMSMapMem-ing, not before you allocate. in TSRs, EMSSave
'       would go in the repetitive bit, while the allocate code would be
'       done just once.
'-----------------------------------------------------------------------
'EXPLANATION:
'The EMM has it's own internal memory block for saving. This command
'calls on it to grab the current EMS status and save. Don't perform
'this function straight at the start of a TSR, but check first if EMS
'is loaded with EMSReady. ems func 47 save page map is used.

FUNCTION EMSSave (EHandle AS WORD) STATIC PUBLIC AS BYTE
DIM EH AS WORD
EH=EHandle
!MOV AH,&h47
!MOV DX,EH
!INT &h67
!CMP AH,0
!JNZ EMSSaveError
EMSSave=%TRUE
EXIT FUNCTION
EMSSaveError:
EMSSave=%FALSE
END FUNCTION

'-----------------------------------------------------------------------
'FUNCTION EMSLoad
'inputs: Your EMS Handle. (WORD)
'outputs: true (-1) if it worked out, false (0) if it didnt. (BYTE)
'notes: See EMSSave for more info.
'-----------------------------------------------------------------------
'EXPLANATION:
'ems func 48 load page map is used. see EMSSave

FUNCTION EMSLoad (EHandle AS WORD) STATIC PUBLIC AS BYTE
DIM EH AS WORD
EH=EHandle
!MOV AH,&h48
!MOV DX,EH
!INT &h67
!CMP AH,0
!JNZ EMSLoadError
EMSLoad=%TRUE
EXIT FUNCTION
EMSLoadError:
EMSLoad=%FALSE
END FUNCTION